/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

// MultiColumn
// A List box that supports multiple columns.
import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;

/*
   * MultiColumn methods

        o MultiColumn(String t[])
          Create a new multi-column list, with the given array of strings as
          column titles. When a list item is selected or double-clicked, an
          event is sent to the parent component with the even argument set
          to "Single" or "Double" respectively.

        o MultiColumn(String t[], MultiColumnCallback c)
          Create a new multi-column list, with the given array of strings as
          titles. When a list item is selected or double-clicked, the
          singleClick or doubleClick methods of the given callback object
          are called, respectively.

        o void addItem(String item[])
          Add one row to the list. The given array should contain the items
          to be added to each column of the new row.

        o void addItems(String item[][])
          Add several rows to the list. The first array dimension is the
          number of rows to add, while the second must be the number of
          columns.

        o String []getItem(int n)
          Returns the elements making up the given row of the list, in an
          array whose length is the number of columns in the list.

        o int selected()
          Returns the number of the currently selected row, or -1

        o void deleteItem(int n)
          Remove the numbered row from the list.

        o void clear()
          Remove all rows from the list.

   * MultiColumnCallback methods

        o void singleClick(MultiColumn list, int num)
          If a MultiColumn object is created with a callback object, this
          method of the callback will be called when a list row is clicked
          on. The list parameter gives the calling list object, and the num
          parameter the number of the chosen row.

        o void doubleClick(MultiColumn list, int num)
          Like the singleClick, but called when a list row is
          double-clicked. When the user double-clicks on a row, both
          singleClick and this method will be called.


*/

public class MultiColumn extends Panel
{
	MultiColumnCallback callback;	// what to call back to 
	String title[];			// column titles
	boolean adjustable = true;
	Color colors[][] = null;
	boolean enabled = true;
	int cpos[];			// column x positions
	float cwidth[];			// proportional column widths
	Vector list[];			// columns of the list
	Scrollbar sb;			// scrollbar at the right side
	int width, height;		// size, minus the scrollbar
	Insets in;			// used space around the border
	int sbwidth;			// width of the scrollbar
	int th;				// height of title bar
	Image bim;			// backing image
	Graphics bg;			// backing graphics
	FontMetrics fnm;		// drawing font size
	int coldrag = -1;		// column being resized
	int sel = -1;			// selected row
	int top = 0;			// first row displayed
	int old_visible = 0;            // num rows visible last time rows
                                        //     were computed
	long last;			// last mouse click time

	static final String defFontName = "helvetica";
	static final int    defFontSize = 12;

	// Create a new list with the given column titles
	MultiColumn(String t[])
	{
	super();
	title = new String[t.length];
	for(int i=0; i<t.length; i++)
		title[i] = t[i];
	list = new Vector[t.length];
	for(int i=0; i<t.length; i++)
		list[i] = new Vector();
	cwidth = new float[t.length];
	for(int i=0; i<t.length; i++)
		cwidth[i] = 1.0f/t.length;
	cpos = new int[t.length+1];
	setLayout(null);
	sb = new Scrollbar(Scrollbar.VERTICAL, 0, 1, 0, 0);
	add(sb);
	validate();
	}

	// Create a new list that calls back to the given object on
	// single or double clicks.
	MultiColumn(String t[], MultiColumnCallback c)
	{
	this(t);
	callback = c;
	}

	// addItem
	// Add a row to the list
	void addItem(String item[])
	{
	for(int i=0; i<title.length; i++)
		list[i].addElement(item[i]);
	repaint();
	compscroll();
	}

	// addItems
	// Add several rows to the list
	void addItems(String item[][])
	{
	for(int i=0; i<item.length; i++)
		for(int j=0; j<title.length; j++)
			list[j].addElement(item[i][j]);
	repaint();
	compscroll();
	}
	
	// addItems
	// Add several row to the list from a Vector
	void addItems(Vector items)
	{
		Enumeration e = items.elements();
		while(e.hasMoreElements())
		{
			String[] item = (String[])e.nextElement();
			for(int j=0; j<title.length; j++)
				list[j].addElement(item[j]);
		}
		repaint();
		compscroll();
	}	
			

	// getItem
	// Returns the contents of a given row
	String []getItem(int n)
	{
	String r[] = new String[title.length];
	for(int i=0; i<title.length; i++)
		r[i] = (String)list[i].elementAt(n);
	return r;
	}

	// selected
	// Return the selected row
	int selected()
	{
	return sel;
	}

	// deleteItem
	// Remove one row from the list
	void deleteItem(int n)
	{
	for(int i=0; i<title.length; i++)
		list[i].removeElementAt(n);
	if (n == sel) sel = -1;
	repaint();
	compscroll();
	}

	// clear
	// Remove everything from the list
	void clear()
	{
	for(int i=0; i<title.length; i++)
		list[i].removeAllElements();
	sel = -1;
	repaint();
	sb.setValues(0, 1, 0, 0);
	}

	// sometimes used with 'clear()'
        void erase_titles()
	{
	for (int i = 0; i < title.length; i++)
		title[i] = "";
	}

	// setWidths
	// Set the proportional widths of each column
	void setWidths(float w[])
	{
	for(int i=0; i<title.length; i++)
		cwidth[i] = w[i];
	respace();
	repaint();
	}

	/**Turns on or off the user's ability to adjust column widths
	 * @param a	Can adjust or not?
	 */
	void setAdjustable(boolean a)
	{
	adjustable = a;
	}

	/**Sets the array of colors used to draw text items.
	 * @param c	The color array (in row/column order), or null to
	 *		use the default
	 */
	void setColors(Color c[][])
	{
	colors = c;
	repaint();
	}

	public void enable()
	{
	enabled = true;
	sb.enable();
	repaint();
	}

	public void disable()
	{
	enabled = false;
	sb.disable();
	repaint();
	}

	// reshape
	// Called when this component gets resized
	public void reshape(int nx, int ny, int nw, int nh)
	{
	if (nw != width+sbwidth || nh != height) {
		in = insets();
		sbwidth = sb.minimumSize().width;
		width = nw-sbwidth - (in.left + in.right);
		height = nh - (in.top + in.bottom);
		sb.reshape(width+in.left, in.top, sbwidth, height);
		respace();

		// Force creation of a new backing image and re-painting
		bim = null;
		repaint();
		compscroll();
		}
	super.reshape(nx, ny, nw, nh);
	}

	// respace
	// Compute pixel column widths from proportional widths
	void respace()
	{
	cpos[0] = 0;
	for(int i=0; i<title.length; i++)
		cpos[i+1] = cpos[i] + (int)(width*cwidth[i]);
	}

	// paint
	// Blit the backing image to the front
	public void paint(Graphics g)
	{
	if (ApplType.getIsVersion1_1())
	    super.paint(g);
		 /* object tries to call Container.paint()
                 when inside netscape -- possibly because this is compiled
                 with a 1.1 compiler, so we avoid this.  Component.paint()
                 just calls this paint again, and would be an infinite loop
                 in netscape; so there is no else clause. */

	if (bim == null) {
		// This is the first rendering
	        //System.out.println("w/h = " + width + "/" + height
                //                          + "  " + size());
		bim = createImage(width, height);
		bg = bim.getGraphics();
		bg.setFont (new Font (defFontName, Font.PLAIN, defFontSize));

                /* TEST
		  Font currfont = bg.getFont();  //is it same one we set?
                  System.out.println ("Current font is " + currfont.toString());
                  Font gfont = g.getFont();
                  System.out.println ("G Current font is " + gfont.toString());
                  NOTE:  This test proves (at least for the standalone
                  application) that we are getting the 'bg Graphics' font,
                  not the 'g Graphics' font.
                */
		fnm = bg.getFontMetrics();
		th = fnm.getHeight() + 4;
		render();
		compscroll();
		}
	g.drawImage(bim, in.left, in.top, this);
	}

	// update
	// Called sometime after repaint()
	public void update(Graphics g)
	{
	if (fnm != null) {
		render();
		paint(g);
		}
	}

	// render
	// Re-draw the list into the backing image
	void render()
	{
	int fh = fnm.getHeight(),	// useful font metrics
	    fd = fnm.getDescent(),
	    fa = fnm.getAscent();
	int bot = Math.min(top+rows()-1, list[0].size()-1);

	// Clear title section and list
	bg.setColor(Color.lightGray);
	bg.fillRect(0, 0, width, th);
	bg.setColor(Color.white);
	bg.fillRect(0, th, width, height-th);

	if (enabled) {
		// Mark the selected row
		if (sel >= top && sel <= bot) {
			bg.setColor(Color.lightGray);
			bg.fillRect(0, th+(sel-top)*fh, width, fh);
			}
		}

	// Draw each column
	for(int i=0; i<title.length; i++) {
		int x = cpos[i], w = cpos[i+1]-x-1;

		// Column title
		bg.setColor(Color.white);
		bg.drawLine(x, 0, x+w, 0);
		bg.drawLine(x, 1, x+w-1, 1);
		bg.drawLine(x, 0, x, th-1);
		bg.drawLine(x+1, 0, x+1, th-2);
		bg.setColor(Color.black);
		bg.drawLine(x, th-1, x+w, th-1);
		bg.drawLine(x, th-2, x+w-1, th-2);
		bg.drawLine(x+w, th-1, x+w, 0);
		bg.drawLine(x+w-1, th-1, x+w-1, 1);
		int tw = fnm.stringWidth(title[i]);
		if (tw < w-6)
			bg.drawString(title[i], x+(w-tw)/2, th-fd-2);

		// Column items
		bg.setColor(Color.lightGray);
		bg.drawLine(x+w-1, th, x+w-1, height);
		bg.setColor(Color.black);
		bg.drawLine(x+w, th, x+w, height);
		for(int j=top; j<=bot; j++) {
			String s = (String)list[i].elementAt(j);
			while(fnm.stringWidth(s) > w-3)
				s = s.substring(0, s.length()-1);
			if (!enabled)
				bg.setColor(Color.lightGray);
			else if (colors != null)
				bg.setColor(colors[j][i]);
			bg.drawString(s, x+1, th+(j+1-top)*fh-fd);
			}
		}
	}

	// mouseDown
	// Select a list item or a column to drag
	public boolean mouseDown(Event e, int x, int y)
	{
	if (!enabled) {
		return true;
		}
	x -= in.left;
	y -= in.top;
	coldrag = -1;
	if (y < th) {
		// Click in title bar
		for(int i=1; i<title.length; i++)
			if (adjustable && Math.abs(cpos[i] - x) < 3) {
				// clicked on a column separator
				coldrag = i;
				}
		if (coldrag == -1) {
			// if not dragging, reset query variables
			sel = -1;
			repaint();
			callback.singleClick(this, -1, -1);
			}
		}
	else {
	  // Item chosen from list
	  int row = (y-th)/fnm.getHeight() + top;
	  int col = 0;                             // added by Heflin

	  // for statement added by Heflin on 5/21/98
	  for (int i=0; i < title.length; i++) {
	    if (x > cpos[i] && x < cpos[i+1])
	      col = i;
	  }

	  if (row < list[0].size()) {
	    // Double-click?
	    boolean dclick = false;
	    if (e.when-last < 1000 && sel == row)
	      dclick = true;
	    else
	      last = e.when;
	    
	    sel = row;
	    repaint();
	    if (callback != null) {
	      // Callback the right function
	      // col argument added by Heflin
	      if (dclick)
		callback.doubleClick(this, row, col);
	      else
		callback.singleClick(this, row, col);
	    }
	    else {
	      // Send an event
	      getParent().postEvent(new Event(this,
					      Event.ACTION_EVENT,
					      dclick?"Double":"Single"));
	    }
	  }
	}
	return true;
	}

	// mouseDrag
	// If a column is selected, change it's width
	public boolean mouseDrag(Event e, int x, int y)
	{
	if (!enabled) {
		return true;
		}
	x -= in.left;
	y -= in.top;
	if (coldrag != -1) {
		if (x > cpos[coldrag-1]+3 && x < cpos[coldrag+1]-3) {
			cpos[coldrag] = x;
			cwidth[coldrag-1] = (cpos[coldrag]-cpos[coldrag-1]) /
					    (float)width;
			cwidth[coldrag] = (cpos[coldrag+1]-cpos[coldrag]) /
					    (float)width;
			repaint();
			}
		}
	return true;
	}

	// handleEvent
	// Deal with scrollbar movement
	public boolean handleEvent(Event evt)
	{
	if (evt.target == sb) {
		// scrollbar moved
		top = sb.getValue();
		compscroll();
		repaint();
		}
	return super.handleEvent(evt);
	}

	// compscroll
	// Re-compute the size of the scrollbar
	private void compscroll()
	{
	  if (fnm == null)        // font metrics
		return;		// not visible
	  int r = rows();
	  //int c = list[0].size() - r;
                //scrollbar not sizing right with this calc
	  int c = list[0].size();
          // might need to change top.  (NEW)
	  int rowchg = r - old_visible;
          if (rowchg > 0)    //might need a new top
          {
	      int rowsAtBottom = list[0].size() - (top + old_visible);
	      if (rowchg > rowsAtBottom)
		  top -= (rowchg - rowsAtBottom);
	      if (top < 0)    //first row is zero (min value)
		  top = 0;
          }
	  sb.setValues(top, r==0?1:r, 0, c<0?0:c);
          old_visible = r;
	}

	// rows
	// Returns the number of rows visible in the list
	private int rows()
	{
	/* TEST
	System.out.println ("height == " + height + " ; fnm.getHeight() == "
                 + fnm.getHeight() + "\n   height / fnm.getHeight() == " +
                 (height / fnm.getHeight()) + " ; list[0].size() == " +
                 list[0].size());
          */

	return Math.min (height / fnm.getHeight() - 1, list[0].size());
	    //
            //list[0].size() == total number of rows in list (used if all
	    //                    are showing)
	    //height / fnm.getHeight() - 1 == number of visible rows
	}
	
	public Dimension minimumSize()
	{
		return new Dimension(100,50);
	}
	
	public Dimension preferredSize()
	{
		return new Dimension(200,100);
	}
}
